/** This file is part of Reactor.
 *  Copyright (c) 2021,2022 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

const version = 22325;

const Logger = require( "server/lib/Logger" );
Logger.getLogger( 'MQTTController', 'Controller' ).always( "Module MQTTController v%1", version );

const Configuration = require("server/lib/Configuration");
const dumpdir = Configuration.getConfig( "reactor.logsdir" );

const Controller = require("server/lib/Controller").requires( 22304 );
const Entity = require("server/lib/Entity").requires( 22306 );

const util = require("server/lib/util");
const isUndef = util.isUndef;

const lexp = require('common/lexp');

// Ref: https://github.com/mqttjs/MQTT.js#readme
const mqtt = require( 'mqtt' );

const { hostname, EOL } = require( 'os' );

const DEFAULT_CONNECTTIMEOUT = 10000;
const DEFAULT_RECONNECTPERIOD = 5000;

var impl = false;
var expr_ce = {};

module.exports = class MQTTController extends Controller {
    constructor( struct, id, config ) {
        super( struct, id, config );

        this.stopping = false;
        this.my_ident = ( this.config.ident || id ).replace( /\s+/g, "" ).replace( /[^a-z0-9_-]/ig, "_" );
        this.topic_prefix = `reactor/${this.my_ident}/`;
        this.log.info( "instance topic ident is %1", this.my_ident );
        this.client = false;
        this.connectTimer = false;
        this.retries = 0;

        this.events = {};
        this.inits = [];
        this.ext_patterns = [];

        this.config.entities = this.config.entities || {};
        this.config.echo = util.coalesce( this.config.echo, false );
        this.config.publish_entities = util.coalesce( this.config.publish_entities, this.config.echo );
        this.config.publish_rule_states = util.coalesce( this.config.publish_rule_states, this.config.echo );
        this.config.publish_global_variables = util.coalesce( this.config.publish_global_variables, this.config.echo );

        this.echo_qos = util.coalesce( parseInt( ( this.config.echo || {} ).qos ), 0 );
        this.echo_retain = !! ( this.config.echo || {} ).retain;
    }

    /** Start the controller.
     */
    async start() {
        /* Load implementation data if not yet loaded */
        if ( !impl ) {
            await this._load_implementation();
        }

        this.system.extendCapability( 'x_mqtt' );

        /* Re-initialize instance */
        this.events = {};
        this.inits = [];
        /* We do not remove external subscriptions. A restart of the controller should resume with
         * the subscriptions in place.
         */
        // this.ext_patterns = [];
        this.stopping = false;

        /* Initialize enumerated entities */
        let allEntities = {};
        Object.keys( this.entities ).forEach( id => allEntities[ id ] = true );
        delete allEntities.system;
        delete allEntities.controller_all;
        for ( let eid of Object.keys( this.config.entities || {} ) ) {
            if ( ! Entity.isValidID( eid ) ) {
                this.log.err( "%1 configuration error: invalid entity ID '%2'", this, eid );
                this.sendError( "Configuration error: invalid entity ID {0:q}", eid );
                continue;
            }
            this.log.debug( 5, "%1 configuring entity %2", this, eid );
            let ec = this.config.entities[ eid ];
            let tpl = {};
            if ( ec.uses_template ) {
                if ( null === impl.templates[ ec.uses_template ] ||
                    "object" !== typeof impl.templates[ ec.uses_template ] ) {
                    this.log.err( "%1 entity %2 uses_template=%3, template not defined", this, eid, ec.uses_template );
                    this.sendError( "Entity {0:q} specifies template {1:q} which is not defined.", eid,
                        ec.uses_template );
                    continue;
                }
                tpl = { ...impl.templates[ ec.uses_template ] };  /* NB: deep copy! */
                if ( Array.isArray( tpl.requires ) ) {
                    let missing = [];
                    for ( let nom of tpl.requires ) {
                        if ( isUndef( ec[ nom ] ) ) {
                            missing.push( nom );
                        }
                    }
                    if ( 0 !== missing.length ) {
                        this.log.err( "%1 entity %2 config missing %3 required by template %4", this, eid,
                            missing.join(', '), ec.uses_template );
                        this.sendError( "Configuration for {0:q} is missing the following required template field(s): {1}",
                            eid, missing.join(', '), ec.uses_template );
                        continue;
                    }
                }
            }
            tpl = this._merge( tpl, { ...ec } );  /* entity configuration may override template values */
            this.log.debug( 6, "%1 merged template for %2 is: %3", this, eid, tpl );

            let e = this.findEntity( eid );
            if ( ! e ) {
                this.log.debug( 6, "%1 created new entity for %2", this, eid );
                e = this.getEntity( tpl.type || 'Entity', eid );
            } else {
                e.deferNotifies( true );
            }

            try {
                e.__template = tpl;
                e.markDead( false );
                delete allEntities[ eid ];
                e.setName( tpl.name || eid );
                e.setType( tpl.type || 'Entity' );
                e.extendCapability( 'x_mqtt_device' );
                e.setAttribute( 'x_mqtt_device.online', false );
                for ( let cap of ( tpl.capabilities || [] )) {
                    try {
                        e.extendCapability( cap );
                    } catch ( err ) {
                        this.log.err( "%1 can't extend capability %2 to %3: %4", this, cap, e, err );
                        this.sendError( "Cannot extend capability {0:q} to entity {1:q}: {2}", cap, eid, String( err ) );
                    }
                }
                if ( "string" === typeof tpl.primary_attribute ) {
                    try {
                        if ( ! e.hasAttribute( tpl.primary_attribute ) ) {
                            /* Not yet present; try to define it. */
                            e.setAttribute( tpl.primary_attribute, null );
                        }
                        e.setPrimaryAttribute( tpl.primary_attribute );
                    } catch ( err ) {
                        this.log.warn( "Can't set primary attribute %1 on %2: %3", tpl.primary_attribute, e, err );
                        this.sendWarning( "Can't set primary attribute to {0:q} on {1:q}: {2}", tpl.primary_attribute,
                            eid, String( err ) );
                    }
                }

                /* Subscribe to other events per the template */
                e.__events = {};
                if ( ! isUndef( tpl.events ) ) {
                    if ( null !== tpl.events && "object" !== typeof tpl.events ) {
                        this.log.err( 'Template for %1 has improperly structured "events" section', eid );
                        this.sendError( "Configuration for {0:q} has improperly structured {1:q} section.", eid, "events" );
                    } else {
                        for ( let topic of Object.keys( tpl.events || {} ) ) {
                            const ev = topic.replace( /%([^%]+)%/g, ( m, p ) => {
                                let altstr = "";
                                const k = p.indexOf( ':' );
                                if ( k >= 0 ) {
                                    altstr = p.substring( k + 1 );
                                    p = p.substring( 0, k );
                                }
                                return String( util.coalesce( ec[p], altstr ) );
                            });
                            this.log.debug( 6, "%1 registering %2 for %3", this, eid, topic );
                            let attr = tpl.events[ topic ];
                            if ( "object" === typeof attr && ! Array.isArray( attr ) ) {
                                /* Convert key/value map form to array form for live use */
                                attr = [];
                                for ( let [k,v] of ( Object.entries( tpl.events[ topic ] ) || {} ) ) {
                                    if ( null === v || "object" !== typeof v ) {
                                        attr.push( { attribute: k, value: v } );
                                    } else {
                                        v.attribute = k;
                                        attr.push( v );
                                    }
                                }
                                this.log.debug( 7, "%1 converted key/value map form for %2 from %3 to %4", this, topic,
                                    tpl.events[ topic ], attr );
                            }
                            e.__events[ev] = attr;
                            this.events[ev] = this.events[ev] || [];
                            this.events[ev].push( eid );
                        }
                    }
                }

                /* Subscribe to LWT for this topic, directing to this entity. */
                if ( "string" === typeof tpl?.lwt ) {
                    const lwt = tpl.lwt.replace( /%([^%]+)%/g, ( m, p ) => {
                        let altstr = "";
                        const k = p.indexOf( ':' );
                        if ( k >= 0 ) {
                            altstr = p.substring( k + 1 );
                            p = p.substring( 0, k );
                        }
                        return String( util.coalesce( ec[p], altstr ) );
                    });
                    e.__events[ lwt ] = this.handle_lwt.bind( this );
                    this.events[ lwt ] = this.events[ lwt ] || [];
                    this.events[ lwt ].push( eid );
                }

                /* Action implementations from the template. */
                e.__actions = {};
                if ( ! isUndef( tpl.actions ) ) {
                    if ( null !== tpl.actions && "object" !== typeof tpl.actions ) {
                        this.log.err( 'Template for %1 has improperly structured "actions" section', eid );
                        this.sendError( "Configuration for {0:q} has improperly formed {1:q} section.", eid, "actions" );
                    } else {
                        for ( let cap of Object.keys( tpl.actions || {} ) ) {
                            for ( let act of Object.keys( tpl.actions[cap] || {} ) ) {
                                const key = cap + '.' + act;
                                e.__actions[ key ] = tpl.actions[cap][act];
                            }
                        }
                    }
                }

                if ( "string" === typeof tpl.query ) {
                    this.inits.push( e );
                }

                if ( ! isUndef( tpl.init ) ) {
                    if ( ! this.inits.includes( e ) ) {
                        this.inits.push( e );
                    }
                }
            } catch ( err ) {
                this.log.err( "%1 error initializing entity %2: %3", this, eid, err );
                this.log.exception( err );
            } finally {
                e.deferNotifies( false );
            }
        }

        Object.keys( allEntities ).forEach( id => {
            let e = this.entities[ id ];
            if ( e ) {
                e.markDead();
                this.log.notice( "%1 device %2 no longer available, marking %3 for removal", this, id, e );
                this.sendWarning( "Controller {0} device {1:q} ({2}) no longer exists.", this.getID(), id, e.getName() );
            }
        });
        this.purgeDeadEntities();

        try {
            const AlertManager = require( 'server/lib/AlertManager' );
            this.subscribe( AlertManager );
        } catch ( err ) {
            this.log.err( '%1 unable to subscribe to alerts: %2', this, err );
        }

        return new Promise( ( resolve ) => {
            this.run();

            if ( this.config.echo ) {
                this.getStructure().waitAllControllers( 60000, this ).then( data => {
                    if ( false !== this.config.publish_entities ) {  /* default publish */
                        this.subscribe_entities();
                    }
                    if ( false !== this.config.publish_global_variables ) {
                        this.subscribe_global_vars();
                    }
                    if ( false !== this.config.publish_rule_states ) {
                        this.subscribe_rules();
                    }
                    return data;
                });
            }
            resolve( this );
        });
    }

    async stop() {
        this.stopping = true;
        this.log.notice( "%1 stopping, sending LWT", this );

        this._clearConnectTimer();

        if ( this.client ) {
            await new Promise( resolve => {
                try {
                    this.client.publish( `${this.topic_prefix}LWT`, 'offline', { qos: 1, retain: true }, err => {
                        this.log.notice( "LWT sent; closing broker connection" );
                        this.client.end( true );
                        this.client = false;
                        resolve( err );
                    });
                } catch ( err ) {
                    this.log.exception( err );
                    resolve();
                }
            });
        }

        this.events = {};
        this.inits = [];
        /* We do not remove external subscriptions. A restart of the controller should resume with
         * the subscriptions in place.
         */
        // this.ext_patterns = [];

        return await super.stop();
    }

    run() {
        if ( this.stopping ) {
            return;
        }
        if ( ! this.client ) {
            let self = this;
            let url = self.config.source || "mqtt://127.0.0.1:1883";
            this.log.notice("Connecting to broker %1", url);
            let copt = self.config.options || {};
            if ( self.config.username ) {
                copt.username = self.config.username;
                copt.password = self.config.password || "";
            }
            copt.clientId = copt.clientId || `reactor_${process.env.HOST || hostname()}_${this.my_ident}`;
            copt.reconnectPeriod = util.coalesce( parseInt( copt.reconnectPeriod ), DEFAULT_RECONNECTPERIOD );
            copt.connectTimeout = util.coalesce( copt.connectTimeout, DEFAULT_CONNECTTIMEOUT );
            copt.resubscribe = false;
            copt.will = {
                topic: `${this.topic_prefix}LWT`,
                payload: 'offline',
                qos: 1,
                retain: true,
                properties: {
                    willDelayInterval: 30
                }
            };
            // copt.rejectUnauthorized = ?;  // for ref/doc only, use with mqtts://, user can provide in "options"

            this.log.debug( 5, "%1 connecting to %2 with options %3", self, url, copt );
            /** Not sure if it's openLuup's MQTT broker (likely) or the MQTT client library, but on reconnect, the
             *  session appears to never get completely set up and "connect" never comes. So, set up a timer to
             *  babysit it. Seems like both need a fix in this instance; mqtt should not stall, regardless of what the
             *  broker may do to it.
             */
            self._startConnectTimer( 2 * copt.connectTimeout );
            self.client = mqtt.connect( url, copt );
            self.client.on( "connect", ( rc ) => {
                self._clearConnectTimer();
                self.log.info( "connected; opening subscriptions" );
                self.log.debug( 7, "%1 connect callback(%2)", self, rc );
                self.client.on( "message", ( topic, payload, packet ) => {
                    if ( self.config.log_topics ) {
                        let match = self.config.log_topics.find( el => topic.match( el ) );
                        if ( match ) {
                            try {
                                const { writeFileSync } = require( 'fs' );
                                const { join } = require( 'path' );
                                writeFileSync( join( dumpdir, `${this.getID()}-topics.log` ),
                                    `${new Date().toISOString()} "${topic}" ${payload}${EOL}`, { flag: "as", encoding: "utf-8" } );
                            } catch ( err ) {
                                self.log.err( "%1 failed to log message for log_topics %2: %3", self, topic, err );
                            }
                        }
                    }
                    if ( topic.startsWith( this.topic_prefix ) ) {  /* includes trailing / */
                        self.handle_controller_event( topic, payload, packet );
                        return;
                    }
                    self.log.debug( 6, "%1 message topic=%2, payload=%3", self, topic, payload );
                    /* Dispatch */
                    if ( false && topic.includes( '/discovery/' ) || topic.endsWith( '/announce/' ) ) { /* tasmota, shelly */
                        try {
                            const { writeFileSync } = require( 'fs' );
                            const { join } = require( 'path' );
                            writeFileSync( join( dumpdir, "mqtt-discovery.log" ),
                                payload, { encoding: "utf-8" } );
                        } catch ( err ) {
                            self.log.err( "%1 failed to dump discovery data for %2: %3", self, topic, err );
                        }
                    }
                    if ( self.events[ topic ] ) {
                        for ( let eid of self.events[ topic ] ) {
                            self.log.debug( 6, "%1 dispatching %2 to %3", self, topic, eid );
                            try {
                                self.handle_event( eid, topic, payload.toString( 'UTF-8' ), packet );
                            } catch ( err ) {
                                self.log.err( "%1 event handler failed for %2 on %3: %4", self, topic, eid, err );
                                self.log.exception( err );
                            }
                        }
                    }
                    if ( self.ext_patterns.length > 0 ) {
                        try {
                            self._handle_ext_event( topic, payload.toString( 'UTF-8' ) );
                        } catch ( err ) {
                            self.log.err( "%1 error while handling external event notifications: %2", self, err );
                            self.log.exception( err );
                        }
                    }
                });
                self.client.subscribe( "#", {}, ( err, granted ) => {
                    self.log.debug( 7, "%1 subscribe all returns err=%2, granted=%3", self, err, granted );
                    if ( err ) {
                        throw new Error( `Unable to subscribe to all: ${String(err)}` );
                    }
                    self.log.info( "Sending inits and going online!" );
                    self.online();
                    self.client.publish( `${this.topic_prefix}LWT`, "online", { qos: 1, retain: true } );
                    self.do_inits();
                });
                try {
                    self.publish_alerts();
                } catch ( err ) {
                    self.log.exception( err );
                }
                self.retries = 0;
            });
            self.client.on( "reconnect", () => {
                self.log.notice( "%1 attempting to reconnect", self );
                if ( 0 === ( ++self.retries % 3 ) ) {
                    self.offline();
                    self.log.notice( "%1 too many retries; discarding client", self );
                    self._recycle();
                    return;
                }
                self._startConnectTimer( 2 * copt.connectTimeout );
            });
            self.client.on( "error", ( err ) => {
                self.log.err( "%1 %2", self, err );
                self._clearConnectTimer();
            });
            self.client.on( "offline", () => {
                self.log.notice( "%1 broker offline!", self );
                self._clearConnectTimer();
                if ( false === self.config.reuse_client ) {
                    self.log.notice( "discarding client" );
                    this._recycle();
                } else if ( 0 === ( ++self.retries % 3 ) ) {
                    self.offline();
                    self.log.notice( "%1 too many retries; discarding client", self );
                    self._recycle();
                } else {
                    self.log.notice( "%1 waiting for auto-reconnect", self );
                }
            });
        }
    }

    /* Force close and recycle connection */
    _recycle() {
        try {
            if ( this.client ) {
                this.client.end( true );
            }
        } catch ( err ) {
            this.log.exception( err );
        } finally {
            this._clearConnectTimer();
            this.client = false;
            if ( ! this.stopping ) {
                this.startDelay( util.coalesce( parseInt( this.config?.options?.reconnectPeriod ),
                    DEFAULT_RECONNECTPERIOD ) );
            }
        }
    }

    _startConnectTimer( timeout ) {
        if ( this.connectTimer ) {
            this._clearConnectTimer();
        }
        this.connectTimer = setTimeout( () => {
            this.log.err( "%1 failed to establish connection", this );
            this.connectTimer = false;
            this._recycle();
        }, timeout || 30000 );
    }

    _clearConnectTimer() {
        if ( this.connectTimer ) {
            try {
                clearTimeout( this.connectTimer );
            } catch ( err ) {
                this.log.debug( 5, "%1 clearing connect timer: %2", this, err );
            }
        }
        this.connectTimer = false;
    }

    /** Merge, for entity configuration. Takes two objects. The contents of obj2 are merged into obj1,
     *  with any overlapping keys being superceded by obj2. Honors subkeys and arrays (that is, existing
     *  values in obj1 in sub-objects and -arrays are merged with those in obj2 through the full depth of
     *  the tree.
     */
    _merge( obj1, obj2 ) {
        function mergeObjects( o1, o2 ) {
            for ( let key of Object.keys( o2 ) ) {
                let v1 = o1[key];
                let v2 = o2[key];
                if ( "undefined" === typeof v1 ) {
                    o1[key] = v2;
                } else if ( Array.isArray( v2 ) ) {
                    if ( Array.isArray( v1 ) ) {
                        o1[key] = v1.concat( v2.filter( el => ! v1.includes( el ) ) );
                    } else {
                        throw TypeError( `Incompatible types at ${key}` );
                    }
                } else if ( null !== v2 && "object" === typeof v2 ) {
                    if ( null === v1 ) {
                        o1[key] = v2;
                    } else if ( "object" === typeof v1 ) {
                        o1[key] = mergeObjects( o1[key], v2 );
                    } else {
                        throw TypeError( `Incompatible types at ${key}` );
                    }
                } else {
                    o1[key] = v2;
                }
            }
            return o1;
        }
        return mergeObjects( obj1 || {}, obj2 || {} );
    }

    /** Load implementation data from data files. */
    async _load_implementation() {
        impl = await this.loadBaseImplementationData( 'mqtt', __dirname );

        const confdir = Configuration.getConfig( "reactor.confdir" );
        let r = util.loadData( [ "." ], "local_mqtt_devices", confdir );
        if ( r ) {
            this.log.logp( 8, this, '_load_implmap', "local templates %1", Object.keys( r.templates || {} ) );
            impl.templates = util.combine( impl.templates || {}, r.templates || {} );
            r = undefined;
        }
    }

    do_device_inits( entity ) {
        let eid = entity.getID();
        let ec = this.config.entities[ eid ];
        let tpl = entity.__template;
        let q = [];
        if ( isUndef( tpl.init ) ) {
            return;
        } else if ( "string" === typeof tpl.init ) {
            q.push( { topic: tpl.init } );
        } else if ( Array.isArray( tpl.init ) ) {
            tpl.init.forEach( el => {
                if ( "string" === typeof el ) {
                    q.push( { topic: el } );
                } else if ( null !== el && "object" === typeof el && "string" === typeof el.topic ) {
                    q.push( el );
                } else {
                    this.log.err( 'Ignoring init for %1 because it cannot be parsed: (%2) %3', entity, typeof el, el );
                }
            });
        } else {
            this.log.err( 'Ignoring invalid "init" data for %1: (%2) %3', entity, typeof tpl.init, tpl.init );
        }
        this.log.debug( 5, "%1 sending inits for %2", this, eid );
        let promises = [];
        for ( let el of ( q || [] ) ) {
            const s = (el.topic || "").replace( /%([^%]+)%/g, ( m, p ) => {
                let altstr = "";
                const k = p.indexOf( ':' );
                if ( k >= 0 ) {
                    altstr = p.substring( k + 1 );
                    p = p.substring( 0, k );
                }
                return String( util.coalesce( ec[p], altstr ) );
            });
            this.log.debug( 6, "%1 init for %2, publishing %3 payload %4", this, eid, s, el.payload );
            let payload = el.payload;
            if ( ! isUndef( payload ) ) {
                if ( 'json' === el.type || ( isUndef( el.type ) && "object" === typeof payload ) ) {
                    payload = JSON.stringify( payload );
                } else {
                    payload = String( payload );
                }
            }
            let opts = {
                qos: util.coalesce( parseInt( el.qos ), 0 ),
                retain: !!el.retain
            };
            let p = new Promise( ( resolve, reject ) => {
                this.client.publish( s, payload, opts, err => {
                    if ( err ) {
                        reject( err );
                    } else {
                        resolve();
                    }
                });
            });
            promises.push( p );
        }
        Promise.allSettled( promises ).then( () => {
            this.log.debug( 6, "%1 all inits for %2 settled", this, eid );
            if ( tpl.query ) {
                entity.perform( 'x_mqtt_device.poll', {} );
            }
        });
    }

    /* Send any inits for configured devices. */
    do_inits() {
        return new Promise( ( resolve ) => {
            for ( let entity of this.inits ) {
                try {
                    this.do_device_inits( entity );
                } catch( err ) {
                    this.log.err( "%1 failed to query %2: %3", this, entity, err );
                    this.log.exception( err );
                }
            }
            resolve();
        });
    }

    handle_event( eid, topic, payload, packet ) {
        this.log.debug( 5, "%1 handling %2 for %3: %4", this, topic, eid, payload );
        let e = this.findEntity( eid );
        if ( ! e ) {
            this.log.err( "%1 can't handle %2 for %3: entity does not exist", this, topic, eid );
            return;
        }
        let ev = e.__events[ topic ];
        if ( ! ev ) {
            this.log.err( "%1 missing event %2 on %3", this, topic, eid );
            return;
        }
        if ( "function" === typeof( ev ) ) {
            ev( e, topic, payload, packet );
            return;
        }
        let ctx = lexp.get_context( e.__template );  /* template items are in global scope */
        lexp.define_var( ctx, 'entity', e.extract() );
        lexp.define_var( ctx, 'system', this.system.extract() );
        lexp.define_var( ctx, 'payload_raw', payload );
        lexp.define_var( ctx, 'packet', packet || {} );
        this.log.debug( 6, "%1 template in context is %2", this, e.__template );
        e.deferNotifies( true );
        try {
            this.log.debug( 6, "%1 events for %2: %3", this, e, ev );
            for ( let d of ev.values() ) {
                let attr = d.attribute;
                let newval = null;
                this.log.debug( 7, "%1 entity %2 attribute %3 source %4", this, eid, attr, d);
                if ( true === d.json_payload ) {
                    try {
                        lexp.define_var( ctx, 'payload', JSON.parse( payload ) );
                    } catch ( err ) {
                        this.log.err( "%1 entity %2 received topic %3 unparsable JSON payload: %4", this, e, topic,
                            err );
                        this.log.err( "%1 payload: %2", this, payload );
                        lexp.define_var( ctx, 'payload', null );
                    }
                } else {
                    lexp.define_var( ctx, 'payload', String( payload ) );
                }
                if ( d.if_expr ) {
                    /* if_expr: conditional expression to determine if this attribute should be modified. This is
                    *  typically used to determine if the MESSAGE contains necessary data.
                    */
                    let ce = expr_ce[ d.if_expr ];
                    if ( isUndef( ce ) ) {
                        try {
                            ce = expr_ce[ d.if_expr ] = lexp.compile( d.if_expr );
                        } catch ( err ) {
                            this.log.err( "%1 can't compile if_expr %2: %3 (%4 %5)", this, d.if_expr, err, eid, topic );
                            ce = expr_ce[ d.if_expr ] = false;
                        }
                    }
                    if ( ce ) {
                        try {
                            let res = lexp.run( ce, ctx );
                            if ( "boolean" !== typeof res ) {
                                this.log.warn( "%1 if_expr expression %2 returned non-boolean (%3)", this, d.if_expr,
                                    typeof res );
                            }
                            if ( ! res ) {
                                continue; /* if_expr returns false; skip this attribute */
                            }
                        } catch ( err ) {
                            this.log.err( "%1 if_expr expression %2 failed evaluation: %3", this, d.if_expr, err );
                            continue; /* hard error evaluating is always false */
                        }
                    } else {
                        continue; /* hard error compiling if_expr is always false */
                    }
                }
                if ( ! isUndef( d.value ) ) {
                    newval = d.value;
                } else {
                    newval = payload;
                }
                if ( d.expr ) {
                    let ce = expr_ce[ d.expr ];
                    if ( isUndef( ce ) ) {
                        try {
                            lexp.define_var( ctx, 'value', newval );
                            ce = expr_ce[ d.expr ] = lexp.compile( d.expr );
                        } catch ( err ) {
                            this.log.err( "%1 can't compile expression %2: %3 (%4 %5)", this, d.expr, err, eid, topic );
                            ce = expr_ce[ d.expr ] = false;
                        }
                    }
                    if ( ce ) {
                        /* Make current value available like in other Controllers? So far, no need. */
                        try {
                            newval = lexp.run( ce, ctx );
                        } catch ( err ) {
                            this.log.err( "%1 expression %2 failed evaluation: %3", this, d.expr, err );
                            newval = null;
                        }
                    }
                }
                if ( isUndef( newval ) ) {
                    newval = null;
                }
                this.log.debug( 5, "%1 setting %2 %3=%4", this, e, attr, newval );
                e.setAttribute( attr, newval );
            }
        } finally {
            /* Assume device is operating */
            e.setAttribute( 'x_mqtt_device.online', true );
            e.deferNotifies( false );
        }
    }

    handle_lwt( e, topic, payload, packet ) {
        this.log.debug( 6, "%1 received LWT (%2) for %3", this, payload, e );
        e.setAttribute( "x_mqtt_device.online", false );
    }

    match_entity( patterns, eid ) {
        for ( let p of ( patterns || [] ) ) {
            if ( p instanceof RegExp ) {
                if ( ! p.test( eid ) ) {
                    continue;
                }
            } else if ( p.startsWith( '/' ) ) {
                /* Regex */
                const k = p.lastIndexOf( '/' );
                const fl = p.substring( k + 1 );
                try {
                    const re = new RegExp( p.substring( 1, k ), fl );
                    if ( ! re.test( eid ) ) {
                        continue;
                    }
                } catch ( err ) {
                    this.log.err( "Failed regular expression %1: %2", p, err );
                }
            } else {
                let m = p.split( '>' );
                let n = eid.split( '>' );
                if ( '*' !== m[0] && m[0] !== n[0] ) {
                    continue;
                }
                if ( '*' !== m[1] && m[1] !== n[1] ) {
                    continue;
                }
            }
            return p;
        }
        return false;
    }

    match_capability( patterns, cap ) {
        for ( let p of ( patterns || [] ) ) {
            if ( p instanceof RegExp ) {
                if ( ! p.test( cap ) ) {
                    continue;
                }
            } else if ( p.startsWith( '/' ) ) {
                /* Regex */
                const k = p.lastIndexOf( '/' );
                const fl = p.substring( k + 1 );
                try {
                    const re = new RegExp( p.substring( 1, k ), fl );
                    if ( ! re.test( cap ) ) {
                        continue;
                    }
                } catch ( err ) {
                    this.log.err( "Failed regular expression %1: %2", p, err );
                }
            } else if ( p !== cap ) {
                continue;
            }
            return p;
        }
        return false;
    }

    match_group( groups, e ) {
        let struct = this.getStructure();
        for ( let gid of ( groups || [] ) ) {
            let group = struct.findEntity( gid );
            if ( group && group.hasMember( e ) ) {
                return true;
            }
        }
        return false;
    }

    subscribe_entities() {
        let struct = this.getStructure();
        this.log.notice( "Subscribing to echo entities" );
        const config = this.config.echo;
        let count = 0;
        for ( let cid of Object.keys( struct.getControllers() ) ) {
            let controller = struct.getControllerByID( cid );
            if ( controller === this ) {
                continue; /* don't self-subscribe */
            }
            let entities = controller.getEntities();
            for ( let [ eid, entity ] of Object.entries( entities ) ) {
                const canonid = entity.getCanonicalID();
                if ( config.include_entities && ! this.match_entity( config.include_entities, canonid ) ) {
                    this.log.debug( 7, "%1 entity %2 not on include_entities list; skipped", this, canonid );
                    continue;
                }
                if ( config.exclude_entities && this.match_entity( config.exclude_entities, canonid ) ) {
                    this.log.debug( 7, "%1 entity %2 is on exclude_entities list; skipped", this, canonid );
                    continue;
                }
                this.log.debug( 6, "%1 echo function: subscribing to %2", this, eid );
                this.subscribe( entity );
                ++count;
            }
        }
        this.log.info( "Subscribed to %1 entities for echo", count );
    }

    notify_mqtt( event ) {
        if ( ! ( this.config.echo && this.client ) ) {
            return;
        }
        const entity = event.sender;
        const controller = entity.getController();
        if ( this === controller ) {
            /* No self-echo */
            return;
        }
        this.log.debug( 6, "%1 echo handling entity notification for %2", this, entity );
        this.echo_entity( entity );
    }

    echo_entity( entity ) {
        this.log.debug( 6, "%1 echo publishing capabilities for entity %2", this, entity );
        let config = this.config.echo;
        const cid = entity.getController().getID();
        const eid = entity.getID();

        /* Check groups, which may be dynamic. */
        if ( config.include_groups && ! this.match_group( config.include_groups ) ) {
            this.log.debug( 7, "%1 entity %2 not in include_groups list; skipped", this, canonid );
            return;
        }
        if ( config.exclude_groups && this.match_group( config.exclude_groups ) ) {
            this.log.debug( 7, "%1 entity %2 is in exclude_groups group; skipped", this, canonid );
            return;
        }

        let etop = eid;
        if ( "name" === config.entity_identifier ) {
            etop = entity.getName().replace( /(\/|\p{C})/gu, "?" );  /* Remove / and Unicode control chars */
        } else if ( "combined" === config.entity_identifier || "id/name" === config.entity_identifier ) {
            etop = `${eid}/${entity.getName().replace( /(\/|\p{C})/gu, "?" )}`;
        } else if ( "name/id" === config.entity_identifier ) {
            etop = `${entity.getName().replace( /(\/|\p{C})/gu, "?" )}/${eid}`;
        } else if ( "id" !== ( config.entity_identifier || "id" ) ) {
            this.log.warn( "%1 configuration entity_identifier unrecognized value: %2", this, config.entity_identifier );
        }

        /* Check capabilities */
        /* ??? extend syntax to include attribute, so specific attributes can be included or excluded? */
        let capabilities = entity.getCapabilities();
        let sent = false;
        for ( let cap of capabilities ) {
            if ( config.include_capabilities && ! this.match_capability( config.include_capabilities, cap ) ) {
                /* Not on list */
                this.log.debug( 6, "%1 echo cap %2 rejected -- not included", this, cap );
                continue;
            }
            if ( config.exclude_capabilities ) {
                if ( this.match_capability( config.exclude_capabilities, cap ) ) {
                    /* On exclude list */
                    this.log.debug( 6, "%1 echo cap %2 rejected -- excluded", this, cap );
                    continue;
                }
            } else if ( this.match_capability( [ /^x_/ ], cap ) ) {
                this.log.debug( 6, "%1 echo cap %2 rejected -- default reject extended %2", this, cap );
                continue;
            }

            /* Build payload for this capability and send data */
            let payload = {};
            const attrs = entity.enumAttributes( cap );
            if ( attrs.length > 0 ) {
                for ( let attr of attrs ) {
                    let an = attr.split( /\./ ).pop();
                    try {
                        payload[an] = entity.getAttribute( attr );
                    } catch ( err ) {
                        payload[an] = null;
                    }
                }
                payload = JSON.stringify( payload );
                const topic = `${this.topic_prefix}${cid}/${etop}/state/${cap}`;
                this.log.debug( 6, "%1 echo publishing %2 %3", this, topic, payload );
                this.client.publish( topic, payload, { qos: this.echo_qos, retain: this.echo_retain } );
                sent = true;
            }
        }
        if ( false !== config.primary_attribute ) {
            const topic = `${this.topic_prefix}${cid}/${etop}/value`;
            const payload = JSON.stringify( util.coalesce( entity.getPrimaryValue() ) );
            this.log.debug( 6, "%1 echo publishing primary attribute topic %2 %3", this, topic, payload );
            this.client.publish( topic, payload, { qos: this.echo_qos, retain: this.echo_retain } );
            sent = true;
        }
        if ( ! sent ) {
            /* Nothing sent; no capabilities included or all excluded. Unsubscribe from entity. */
            this.log.notice( "Removing %1 from echo watch because it has no publishable capabilities", entity );
            this.unsubscribe( entity );
        }
    }

    publish_alerts( manager ) {
        if ( this.client ) {
            manager = manager || require( 'server/lib/AlertManager' );
            const alerts = manager.getExportData();  /* Safe form/content */
            const topic = `${this.topic_prefix}alerts`;
            this.log.debug( 6, "%1 publishing alerts topic %2 %3", this, topic, alerts );
            this.client.publish( topic, JSON.stringify( alerts ), { qos: 1, retain: true } );
        }
    }

    notify( event ) {
        try {
            switch ( event.type ) {
                case 'entity-changed':
                    this.notify_mqtt( event );
                    break;
                case 'alerts-changed':
                    this.publish_alerts( event.sender );
                    break;
                default:
                    /* ignored */
            }
        } catch ( err ) {
            this.log.err( "%1 exception while handling %2 event from %3: %4", this, event.type, event.sender, err );
            this.log.exception( err );
        }

        return super.notify( event );  /* REQUIRED! */
    }

    /* Handle a controller event for an echo device. Incoming 'perform' messages are translated into
    *  entity and action, with the message payload being the action arguments. This allows pretty
    *  much any action in any capability known to any entity to be performed via MQTT.
    */
    handle_controller_event( topic, payload, packet ) {
        if ( Buffer.isBuffer( payload ) ) {
            payload = payload.toString( 'UTF-8' );
        } else if ( null === payload ) {
            payload = "";
        }
        let m = topic.split( '/' );
        if ( m.length < 3 || m[1] !== this.my_ident ) {
            return;  /* Not for me! */
        }
        this.log.debug( 6, "%1 handling incoming %2 %3", this, topic, payload );
        if ( "ping" === m[2] ) {
            /* reactor/mqtt/ping */
            this.client.publish( `${this.topic_prefix}pong`, payload || Date.now(), { qos: 0, retain: false } );
            return;
        }
        if ( "Rule" === m[2] ) {
            if ( "state" === m[4] ) {
                return;
            } else if ( "query" === m[4] ) {
                /* reactor/mqtt/Rule/rule-j29z3yt/query */
                const Rule = require( "server/lib/Rule" );
                if ( Rule.exists( m[3] ) ) {
                    try {
                        this.publish_rule( m[3], {} );
                    } catch ( err ) {
                        this.log.exception( err );
                    }
                } else {
                    this.log.warn( "%1 topic %2: rule does not exist", this, topic );
                }
            } else {
                this.log.warn( "%1 invalid topic %2 (ignored)", this, topic );
            }
            return;
        }
        if ( "Expr" === m[2] ) {
            /* reactor/mqtt/Expr/somevarname/query (payload ignored) */
            /* reactor/mqtt/Expr/somevarname/set payload=value */
            if ( "value" === m[4] ) {
                return;
            } else if ( "query" === m[4] ) {
                /* Echo value */
                this.publish_global_var( m[3], {} );
            } else if ( "set" === m[4] ) {
                const Engine = require( "server/lib/Engine" );
                /* Set value */
                let value = null;
                try {
                    value = util.coalesce( JSON.parse( payload ) );
                } catch ( err ) {
                    this.log.err( "%1 unparsable JSON payload in %2 %3", this, topic, payload );
                }
                try {
                    this.log.debug( 5, "%1 handling %2 set %3=%4", this, topic, m[3], value);
                    // Engine.getInstance().setGlobalVariable( m[3], value );  /* requires Engine 22029 */
                    Engine.getInstance().setVariable( m[3], false, value, false, false );
                } catch ( err ) {
                    this.log.err( "%1 topic %2 can't set global variable %3: %4", this, topic, m[3], err );
                    this.log.exception( err );
                }
            } else {
                this.log.warn( "%1 invalid topic %2 (ignored)", this, topic );
            }
            return;
        }
        if ( 5 === m.length && "query" === m[4] ) {
            /* reactor/mqtt/vera/device_10/query */
            let eid = m[2] + '>' + m[3];
            const entity = this.getStructure().findEntity( eid );
            if ( entity ) {
                this.echo_entity( entity );
            } else {
                this.log.warn( "received topic %1 for entity %2, no such entity (ignored)", topic, eid );
                return;
            }
        }
        /* Ex: reactor/my_ident/vera/device_10/perform/power_switch/on -- 7 parts */
        if ( m.length < 5 || "perform" !== m[4] ) {
            /* Not something we're handling here. The "state" and "value" msg echoes are also caught here. */
            this.log.debug( 7, "%1 quietly ignoring %2", this, topic );
            return;
        }
        let eid = `${m[2]}>${m[3]}`;
        let entity = this.getStructure().findEntity( eid );
        if ( ! entity ) {
            this.log.notice( "received topic %1 for entity %2, no such entity (ignored)", topic, eid );
            return;
        }
        if ( ! entity.hasCapability( m[5] ) ) {
            this.log.notice( "topic %1 for entity %2: entity does not support capability %3", topic, eid, m[5] );
            return;
        }
        let action = m[5] + '.' + m[6];
        let parameters = {};
        if ( "undefined" !== typeof payload && null !== payload && "" !== payload ) {
            try {
                parameters = JSON.parse( payload );
            } catch ( err ) {
                this.log.warn( "%1 echo topic %2 invalid JSON: %3", this, topic, err );
                this.log.info( "payload=%1", payload );
                return;
            }
        }
        this.log.debug( 5, "%1 echo sending action %2 parameters %3", this, action, parameters );
        entity.perform( action, parameters ).then( data => {
            this.log.info( "%1 action %2 on %3 completed (MQTT echo)", this, action, entity );
            return data;
        });
    }

    performOnEntity( entity, actionName, params ) {
        this.log.debug( 5, "%1 perform %2 on %3 with %4", this, actionName, entity, params );
        const eid = entity.getID();
        const cf = this.config.entities[ eid ] || {};
        let act = ( entity.__actions || {} )[ actionName ];
        if ( ! act ) {
            return super.performOnEntity( entity, actionName, params );
        }
        if ( "string" === typeof act ) {
            act = { send: act };
        }

        let ctx = lexp.get_context();
        lexp.define_var( ctx, 'entity', entity.extract() );
        lexp.define_var( ctx, 'system', this.system.extract() );
        lexp.define_var( ctx, 'parameters', params || {} );
        lexp.define_var( ctx, 'config', cf );
        for ( const [key, val] of Object.entries( cf ) ) {
            lexp.define_var( ctx, key, val );
        }

        let topic = ( act.topic || "undefined/topic" ).replace( /%([^%]+)%/g, ( m, p ) => {
            let altstr = "";
            const k = p.indexOf( ':' );
            if ( k >= 0 ) {
                altstr = p.substring( k + 1 );
                p = p.substring( 0, k );
            }
            return String( util.coalesce( cf[p], altstr ) );
        });

        let payload;
        if ( "string" === typeof act.payload ) {
            payload = String( act.payload );
        } else if ( null !== act.payload && "object" === typeof act.payload ) {
            payload = util.coalesce( act.payload.value );
            if ( act.payload.expr ) {
                if ( isUndef( expr_ce[ act.payload.expr ] ) ) {
                    try {
                        expr_ce[ act.payload.expr ] = lexp.compile( act.payload.expr );
                    } catch ( err ) {
                        this.log.err( "%1 entity %2 action %3 payload expression failed: %4", this, eid, actionName, err );
                        expr_ce[ act.payload.expr ] = false;
                    }
                }
                if ( expr_ce[ act.payload.expr ] ) {
                    try {
                        payload = lexp.run( expr_ce[ act.payload.expr ], ctx );
                    } catch ( err ) {
                        this.log.err( "%1 entity %2 action %3 payload expression failed: %4", this, eid, actionName, err );
                    }
                }
            }
            if ( 'json' === act.payload.type ) {
                payload = JSON.stringify( payload );
            } else if ( 'raw' !== ( act.payload.type || 'raw' ) ) {
                this.log.warn( "%1 entity %2 action %3 unsupported payload type %4", this, eid, actionName, act.payload.type );
            }
        }

        let opts = {
            qos: util.coalesce( parseInt( act.qos ), 0 ),
            retain: !!act.retain
        };

        return new Promise( ( resolve, reject ) => {
            this.log.debug( 5, "%1 perform %2 on %3: publishing %4 %5", this, actionName, eid, topic, payload );
            this.client.publish( topic, payload, opts, (err) => {
                if ( err ) {
                    reject( err );
                } else {
                    resolve();
                }
            });
        });
    }

    subscribe_global_vars() {
        const GlobalExpression = require( 'server/lib/GlobalExpression' );
        this.log.info( "%1 subscribing to global variables", this );
        const exps = GlobalExpression.getExpressions();
        for ( let id of exps ) {
            this.log.debug( 5, "%1 subscribing to %2", this, id );
            const exp = GlobalExpression.getInstance( id );
            this.subscribe( exp, this.publish_global_var.bind( this, id ) );
            try {
                this.publish_global_var( id );
            } catch ( err ) {
                this.log.warn( "%1 error while publishing current global %2 value: %3", this, exp.name, err );
            }
        }
    }

    publish_global_var( id ) {
        const GlobalExpression = require( 'server/lib/GlobalExpression' );
        if ( GlobalExpression.exists( id ) ) {
            const exp = GlobalExpression.getInstance( id );
            const payload = JSON.stringify( exp.getValue() );
            const topic = `${this.topic_prefix}Expr/${exp.name}/value`;
            const opts = {
                qos: this.echo_qos,
                retain: this.echo_retain
            };
            this.log.debug( 5, "%1 publishing gvar %2 value %3 %4", this, exp.name, topic, payload.substring( 0, 63 ) );
            this.client.publish( topic, payload, opts, err => {
                if ( err ) {
                    this.log.warn( "MQTT error (%1): %2", topic, String( err ) );
                }
            });
        } else {
            this.log.warn( "%1 can't publish global expr %2: does not exist", this, id );
        }
    }

    subscribe_rules() {
        const Engine = require( "server/lib/Engine" );
        const engine = Engine.getInstance( this.getStructure() );
        const Rulesets = require( "server/lib/Rulesets" );
        const Ruleset = require( "server/lib/Ruleset" );
        const Rule = require( "server/lib/Rule" );
        const rulesets = Rulesets.getInstance().getRulesets();
        for ( let rsid of rulesets ) {
            const ruleset = Ruleset.getInstance( rsid );
            for ( let rid of ruleset.getRules() ) {
                if ( Rule.exists( rid ) ) {
                    const rule = Rule.getInstance( rid, engine );
                    const bound_handler = this.publish_rule.bind( this, rid );
                    this.subscribe( rule, bound_handler );
                    this.subscribe( rule.getRuleStates(), bound_handler );
                    try {
                        this.publish_rule( rid, { type: 'rule-state' } );
                    } catch ( err ) {
                        this.log.err( "%1 failed initial publish of %2: %3", this, rid, err );
                    }
                }
            }
        }
    }

    publish_rule( rid, event ) {
        this.log.debug( 6, "%1 rule %2 event %3", this, rid, event.type );
        if ( "rule-state" !== event.type ) {
            return;
        }
        const Rule = require( "server/lib/Rule" );
        if ( Rule.exists( rid ) ) {
            const Engine = require( "server/lib/Engine" );
            const engine = Engine.getInstance( this.getStructure() );
            const rule = Rule.getInstance( rid, engine );
            const payload = JSON.stringify( util.coalesce( rule.getConditionState( 'rule' ).evalstate ) );
            const topic = `${this.topic_prefix}Rule/${rule.getID()}/state`;
            const opts = {
                qos: this.echo_qos,
                retain: this.echo_retain
            };
            this.log.debug( 5, "%1 publishing rule %2 state %3 %4", this, rule, topic, payload );
            this.client.publish( topic, payload, opts, err => {
                if ( err ) {
                    this.log.warn( "MQTT error (%1): %2", topic, String( err ) );
                }
            });
        } else {
            this.log.warn( "%1 can't publish %2: does not exist", this, rid );
        }
    }

    /** ---------------- EXTERNAL INTERFACES -------------------- */

    /** Allow another controller or plugin to subscribe to MQTT topics.
     *  The callback will be used to respond to a matching topic.
     *  It may be a function, which is called with the topic and payload
     *  as arguments, or a Controller, which is sent a notification of
     *  type 'ext-mqtt-topic' with the topic and payload in data.
     */
    extSubscribeTopic( obj, topic, callback, ...args ) {
        if ( ! ( obj instanceof Controller ) || this.isEqual( obj ) ) {
            /* Must be controller, and don't allow mistaken pass of MQTTController */
            throw new TypeError( "Invalid first argument, must be Controller (this)" );
        }
        if ( "string" !== typeof topic ) {
            throw new TypeError( "Invalid topic" );
        }
        if ( ! ( isUndef( callback ) || "function" === typeof callback ) ) {
            throw new TypeError( "Invalid callback" );
        }
        let pattern = '^' + topic.replace( /[.?*]/g, "\\$&" ) + '$';
        pattern = pattern.replace( /\/\+/g, "/[^/]+" );
        pattern = pattern.replace( /\/#/g, "/.*" );
        let re = new RegExp( pattern );
        this.ext_patterns.push( { topic: topic, obj: obj, re: re, callback: callback, args: args } );
        this.log.debug( 5, "%1 extSubscribeTopic added %2 as %3 for %4 cb=%5", this, topic, re.toString(), obj,
            typeof callback );
    }

    /** Unsubscribe from a previously-subscribed topic. */
    extUnsubscribeTopic( obj, topic ) {
        this.log.debug( "%1 extUnsubscribe %2 for %3", this, topic, obj );
        while ( true ) {
            const ix = this.ext_patterns.findIndex( el => el.obj.isEqual( obj ) && ( !topic || el.topic === topic ) );
            if ( ix < 0 ) {
                break;
            }
            this.log.debug( 5, "%1 extUnsubscribeTopic removing %2 for %3", this, this.ext_patterns[ix].topic,
                this.ext_patterns[ix].obj );
            this.ext_patterns.splice( ix, 1 );
        }
    }

    /** Publish a topic. */
    async extPublishTopic( topic, payload, opts ) {
        return new Promise( ( resolve, reject ) => {
            if ( "object" === typeof payload && ! opts.raw ) {
                payload = JSON.stringify( payload );
            } else if ( ! isUndef( payload ) ) {
                payload = String( payload );
            }
            opts = opts || {};
            opts.qos = util.coalesce( parseInt( opts.qos ), 0 );
            opts.retain = "boolean" === typeof opts.retain && opts.retain;
            this.log.debug( 5, "%1 extPublishTopic() publishing %2 qos %3 retain %4 payload %5", this,
                topic, opts.qos, opts.retain, payload || "" );
            this.client.publish( topic, payload, opts, err => {
                if ( err ) {
                    this.log.warn( "MQTT error (%1): %2", topic, String( err ) );
                    reject( err );
                } else {
                    resolve();
                }
            });
        });
    }

    /** Private method to dispatch topics to handlers */
    _handle_ext_event( topic, payload ) {
        let pt = this.ext_patterns.filter( el => topic.match( el.re ) );
        this.log.debug( 5, "%1 _handle_ext_event() %2 matches %3 subscriptions", this, topic, pt.length );
        for ( let p of pt ) {
            if ( "function" === typeof p.callback ) {
                try {
                    p.callback( topic, payload, ...(p.args || []) );
                } catch ( err ) {
                    this.log.err( "%1 error thrown by external topic %2 handler (callback): %3", this, topic, err );
                    this.log.exception( err );
                }
            } else if ( ! p.callback ) {
                const c = p.obj;
                try {
                    c.notify( {
                        type: 'ext-mqtt-topic',
                        sender: this,
                        currentSender: this,
                        data: { topic: topic, payload: payload, subscribed_as: p.topic },
                        args: p.args || []
                    });
                } catch ( err ) {
                    this.log.err( "%1 error thrown by external topic handler for %2 (in %3 notify()): %4", this,
                        topic, c, err );
                    this.log.exception( err );
                }
            }
        }
    }

    /** ---------------- DEFAULT ACTION IMPLEMENTATIONS ------------------ */

    action_toggle_toggle( entity, params ) {
        return this.performOnEntity( entity, "power_switch.set",
            { state: ! entity.getAttribute( "power_switch.state" ) } );
    }

    action_x_mqtt_publish( entity, params ) {
        this.log.debug( 5, "%1 x_mqtt.publish action: %2", this, params );
        return new Promise( ( resolve, reject ) => {
            let payload = params.payload;
            if ( "" === payload || null === payload ) {
                payload = undefined;
            } else if ( "object" === typeof payload ) {
                payload = JSON.stringify( payload );
            }
            let opts = {
                qos: util.coalesce( parseInt( params.qos ), 0 ),
                retain: !!params.retain
            };
            this.log.debug( 6, "%1 x_mqtt.publish publishing %2 payload %3", this, params.topic, payload );
            this.client.publish( params.topic, payload, opts, err => {
                if ( err ) {
                    this.log.warn( "MQTT error (%1): %2", params.topic, String( err ) );
                    reject( err );
                } else {
                    resolve();
                }
            });
        });
    }

    action_x_mqtt_device_poll( entity, params ) {
        return new Promise( ( resolve, reject ) => {
            let tpl = entity.__template;
            if ( tpl && tpl.query ) {
                let eid = entity.getID();
                let ec = this.config.entities[ eid ] || {};
                const s = tpl.query.replace( /%([^%]+)%/g, ( m, p ) => {
                    let altstr = "";
                    const k = p.indexOf( ':' );
                    if ( k >= 0 ) {
                        altstr = p.substring( k + 1 );
                        p = p.substring( 0, k );
                    }
                    return String( util.coalesce( ec[p], altstr ) );
                });
                this.log.debug( 5, "%1 x_mqtt_device.poll for %2, sending %3", this, entity, s );
                this.client.publish( s, undefined, { qos: 0, retain: false }, err => {
                    if ( err ) {
                        reject( err );
                    } else {
                        resolve();
                    }
                });
            } else {
                reject( new Error( `poll is not implemented for this entity` ) );
            }
        });
    }

};
